# SPDX-FileCopyrightText: 2024 Earth System Data Exploration (ESDE), Jülich Supercomputing Center (JSC)
#
# SPDX-License-Identifier: MIT

"""
Driver-script to train downscaling models when benchmarking with JUBE.
"""

__author__ = "Michael Langguth"
__email__ = "m.langguth@fz-juelich.de"
__date__ = "2023-12-XX"
__update__ = "2024-03-07"

import os, sys
import socket
import argparse
from datetime import datetime as dt
print("Start with importing packages at {0}".format(dt.strftime(dt.now(), "%Y-%m-%d %H:%M:%S")))
import gc
import json as js
from timeit import default_timer as timer
import numpy as np
import xarray as xr
import tensorflow as tf
from tensorflow.keras.utils import plot_model
import horovod.tensorflow.keras as hvd
from all_normalizations import ZScore
from model_engine import ModelEngine
from model_utils import check_horovod, set_gpu_memory_growth
from handle_data_class import prepare_dataset
from other_utils import print_gpu_usage, print_cpu_usage, copy_filelist, get_batch_size_mb
import energy_utils
from benchmark_utils import get_training_time_dict

# Open issues:
# * d_steps must be parsed with hparams_dict as model is uninstantiated at this point and thus no default parameters
#   are available

def main_train(parser_args):
    # start timing
    t0 = timer()

    # check if Horovod is available
    main_process = True
    num_processes = 1
    with_horovod = check_horovod()    
    if with_horovod:
        # get GPU devices and set-up Horovod
        gpus = tf.config.experimental.list_physical_devices("GPU")

        set_gpu_memory_growth()

        hvd.init()

        if len(gpus) == 0:
            raise Exception("No GPUs available")
        if len(gpus) > 1:
            tf.config.experimental.set_visible_devices(gpus[hvd.local_rank()], "GPU")
        main_process = hvd.rank() == 0
        num_processes = hvd.size()
        add_str = f" on GPU {hvd.rank()}"
        seed = 42 + hvd.size()  
        if main_process:
            print(f"Horovod initialized with {num_processes} GPUs.")
            print(f"Seed for random sampling: {seed}")
    else:
        add_str = ""
        seed = None

    # Get some basic directories and flags
    datadir = parser_args.input_dir
    outdir = parser_args.output_dir
    job_id = parser_args.id
    dataset = parser_args.dataset.lower()
    js_norm = parser_args.js_norm

    if main_process:
        print(f"Job ID: {job_id}")
        print(f"Number of GPUs: {num_processes}")

    # initialize checkpoint-directory path for saving the model
    model_savedir = os.path.join(outdir, parser_args.exp_name)

    # read configuration files for model and dataset
    with parser_args.conf_ds as dsf:
        ds_dict = js.load(dsf)

    with parser_args.conf_md as mdf:
        hparams_dict = js.load(mdf)

    # get normalization object if corresponding JSON-file is parsed
    if js_norm:
        data_norm = ZScore(ds_dict["norm_dims"])
        data_norm.read_norm_from_file(js_norm)
        norm_dims, write_norm = None, False
    else:
        data_norm, write_norm = None, True
        norm_dims = ds_dict["norm_dims"]
        if with_horovod:
            raise ValueError("Normalization object must be provided when using Horovod.")
        
    # get model instance and set-up batch size
    model_instance = ModelEngine(parser_args.model)
        
    # get tensoflow dataset objects for training and validation data
    # training dataset
    print(f"Start preparing training data{add_str}...")

    t0_train = timer()
    tfds_train, train_info = prepare_dataset(datadir, dataset, ds_dict, hparams_dict, "train", ds_dict["predictands"], 
                                             norm_obj=data_norm, norm_dims=norm_dims, with_horovod=with_horovod) 
    
    data_norm, shape_in, nsamples, tfds_train_size = train_info["data_norm"], train_info["shape_in"], \
                                                     train_info["nsamples"], train_info["dataset_size"]
    tfds_train_size, tfds_train_size_eff = train_info["dataset_size"], train_info["effective_dataset_size"]
    ds_obj_train = train_info.get("ds_obj", None)

    if write_norm and main_process:
        data_norm.save_norm_to_file(os.path.join(model_savedir, "norm.json"))

    # Tracking training data preparation time
    if main_process:
        ttrain_load = timer() - t0_train
        print(f"TF training Dataset preparation time: {ttrain_load:.2f}s. Training data loaded into memory? {ds_obj_train is None}.")
        
    print(f"Start preparing validation data{add_str}...")
    
    # validation dataset
    t0_val = timer()
    tfds_val, val_info = prepare_dataset(datadir, dataset, ds_dict, hparams_dict, "val", ds_dict["predictands"], 
                                         norm_obj=data_norm, with_horovod=with_horovod) 
    
    ds_obj_val = val_info.get("ds_obj", None)
    
    # Tracking validation data preparation time
    if main_process:
        tval_load = timer() - t0_val
        print(f"TF validation Dataset preparation time: {tval_load:.2f}s. Validation data loaded into memory? {ds_obj_val is None}.")
        print("Finished data preparation")

        # copy configuration and normalization JSON-file to model-directory (incl. renaming)
        os.makedirs(model_savedir, exist_ok=True)
        filelist, filelist_new = [parser_args.conf_ds.name, parser_args.conf_md.name], [f"config_ds_{dataset}.json", f"config_{parser_args.model}.json"]
        if not write_norm:
            filelist.append(js_norm), filelist_new.append(os.path.basename(js_norm))
        
        copy_filelist(filelist, model_savedir, filelist_new)

    # instantiate model and...
    # Note: Parse varnames from train_info since list of varnames might get updated depending on model and dataset configuration
    model = model_instance(shape_in, list(train_info["varnames_tar"]), hparams_dict, model_savedir, parser_args.exp_name, with_horovod=with_horovod) 

    # ... compile
    model.compile(**model.compile_options)

    # train model (note that callbacks are part of the fit_options by default)
    steps_per_epoch = int(np.ceil(nsamples / ds_dict["batch_size"]))

    if main_process:
        hparams = model.hparams
        bs_train = hparams["batch_size"]*(hparams["d_steps"] + 1) if "d_steps" in hparams else hparams["batch_size"]

        print("**************************************************************************")
        print(f"Raw training dataset size: {tfds_train_size / 1.e+06:.2f} MB.")
        if tfds_train_size != tfds_train_size_eff:
            print(f"Effective training dataset size: {tfds_train_size_eff / 1.e+06:.2f} MB.")
        print(f"Batches per epoch: {steps_per_epoch}")  
                                    
        print(f"TF training Dataset preparation time: {timer() - t0_train:.2f}s.")
        print(f"Number of epochs: {hparams['nepochs']}")
        print(f"Samples per batch: {hparams['batch_size']}; Batch size: {get_batch_size_mb(shape_in, hparams['batch_size']):.2f} MB")

        if ds_dict["batch_size"] != bs_train:
            print(f"Batch size per training step: {get_batch_size_mb(shape_in, bs_train):.2f} MB")

        print("**************************************************************************")
        print(f"Start training of {parser_args.model.capitalize()}...")

    # run training
    history = model.fit(x=tfds_train, epochs=model.hparams["nepochs"],
                        steps_per_epoch=steps_per_epoch, validation_data=tfds_val, validation_steps=300,
                        verbose=2, **model.fit_options)

    if main_process:
        # get some parameters from tracked training times and put to dictionary
        training_times = get_training_time_dict(model.fit_options["callbacks"][0].epoch_times,
                                                nsamples * model.hparams["nepochs"])
        
        # print training times
        print(f"Total training time: {training_times['Total training time']:.1f} s")
        tval_load = sum(ds_obj_val.reading_times) if ds_obj_val else tval_load
        ttrain_load = sum(ds_obj_train.reading_times) if ds_obj_train else ttrain_load
        print(f"Data loading time: {ttrain_load + tval_load:.2f}s.")
        if ds_obj_train:
            # estimated throughput for training data if not all data was loaded into memory
            scal_factor = hvd.size() if with_horovod else 1
            print(f"Average throughput: {ds_obj_train.ds_proc_size / 1.e+06 / training_times['Total training time']*scal_factor:.3f} MB/s")

        # plot model architecture
        try:
            if callable(model.__dict__.get("plot_model", None)):
                model.plot_model(model_savedir, show_shapes=True) # , show_layer_actiavtions=True)
            else:
                plot_model(model, os.path.join(model_savedir, f"plot_{parser_args.exp_name}.png"),
                           show_shapes=True) #, show_layer_activations=True)
        except Exception as _:
            print("Model could not be plotted. Skip producing model plot.")

        # save trained model
        t0_save = timer()

        model_savedir_last = os.path.join(model_savedir, f"{parser_args.exp_name}_last")
        os.makedirs(model_savedir, exist_ok=True)
        model.save(filepath=model_savedir_last)


        # final timing
        tend = timer()
        saving_time = tend - t0_save
        print(f"Model saving time: {saving_time:.2f}s")
        
        # get trainable and untrainable parameters
        trainable_params, untrainable_params = model.count_params()
        print(f"# trainable parameters: {trainable_params}, # untrainable parameters: {untrainable_params}")
       
        # end timing
        tot_run_time = tend - t0
        print(f"Total runtime: {tot_run_time:.1f}s")
        # some statistics on memory usage
        print_gpu_usage("Final GPU memory: ")
        print_cpu_usage("Final CPU memory: ")

        print("Finished training at {0}".format(dt.strftime(dt.now(), "%Y-%m-%d %H:%M:%S")))
        print("**************************************************************************")

    return main_process


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_dir", "-in", dest="input_dir", type=str, required=True,
                        help="Directory where input netCDF-files are stored.")
    parser.add_argument("--output_dir", "-out", dest="output_dir", type=str, required=True,
                        help="Output directory where model is savded.")
    parser.add_argument("--downscaling_model", "-model", dest="model", type=str, required=True,
                        help="Name of model architeture used for downscaling.")
    parser.add_argument("--downscaling_dataset", "-dataset", dest="dataset", type=str, required=True,
                        help="Name of dataset to be used for downscaling model.")
    parser.add_argument("--experiment_name", "-exp_name", dest="exp_name", type=str, required=True,
                        help="Name for the current experiment.")
    parser.add_argument("--configuration_model", "-conf_md", dest="conf_md", type=argparse.FileType("r"), required=True,
                        help="JSON-file to configure model to be trained.")
    parser.add_argument("--configuration_dataset", "-conf_ds", dest="conf_ds", type=argparse.FileType("r"),
                        required=True, help="JSON-file to configure dataset to be used for training.")
    parser.add_argument("--json_norm_file", "-js_norm", dest="js_norm", type=str, default=None,
                        help="JSON-file providing normalization parameters.")
    parser.add_argument("--job_id", "-id", dest="id", type=int, required=True, help="Job-id from Slurm.")
    parser.add_argument("--hardware_name", "-hw", dest="hardware_name", type=str, required=True,
                        help="Name of hardware used for training.")

    args = parser.parse_args()

    if args.hardware_name=='MI250_GPU':
        sys.path.append('/opt/rocm/libexec/rocm_smi')

    energy_profiler = energy_utils.get_energy_profiler(args.hardware_name)

    hostname = socket.gethostname().split('.')[0]
    print(f"Host: {hostname} on {args.hardware_name}")

    with energy_profiler() as measured_scope:
        try:
            main_process = main_train(args)
        except Exception as exc:
            import traceback
            print(f"Errors occured during training: {exc}")
            print(f"Traceback: {traceback.format_exc()}")

    if main_process:
        ## Remove data from the first epoch
        ## start_time_second_epoch = timing_callback.start_times[1]
        # measured_scope.df = measured_scope.df[measured_scope.df["timestamps"] >= start_time_second_epoch]

        f = open(f"EnergyFile-NVDA-integrated.txt", 'a')
        measured_scope.df.to_csv("EnergyFile-NVDA.csv")

        print("Energy-per-GPU-list:")
        max_power=measured_scope.df.loc[:,(measured_scope.df.columns != 'timestamps')].max().max()
        print(f"Max Power: {max_power:.2f} W")
        
        max_agg_power=measured_scope.df.loc[:,(measured_scope.df.columns != 'timestamps')].sum(axis=1).max()
        print(f"Max Aggregate Power: {max_agg_power:.2f} W")
        
        mean_agg_power=measured_scope.df.loc[:,(measured_scope.df.columns != 'timestamps')].sum(axis=1).mean()
        print(f"Mean Aggregate Power: {mean_agg_power:.2f} W")
        
        if args.hardware_name=='MI250_GPU':
            energy_int,energy_cnt = measured_scope.energy()
            print(f"Integrated Total Energy: {np.sum(energy_int):.2f} Wh")
            print(f"Counter Total Energy: {np.sum(energy_cnt):.2f} Wh")
            
            f.write(f"integrated: {energy_int}") 
            f.write(f"from counter: {energy_cnt}")
            f.close()
        elif args.hardware_name in ['A100_GPU','V100_GPU','H100_GPU','GC200_IPU', 'A2_GPU', 'GH200_GPU']:
            energy_int = measured_scope.energy()
            print(f"Integrated Total Energy: {np.sum(energy_int):.2f} Wh")
            f.write(f"integrated: {energy_int}") 
            f.close()       

